# noqa
from dodo_commands.system_commands import DodoCommand
import os


class Command(DodoCommand):  # noqa
    help = (
        "Writes (or removes) a small script that activates the latest " +
        "Dodo Commands project"
    )
    safe = False

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('status', choices=['on', 'off'])

    def handle_imp(self, status, **kwargs):  # noqa
        script = os.path.expanduser("~/.dodo_commands_autostart")
        if status == "on" and not os.path.exists(script):
            with open(script, "w") as f:
                f.write("$(dodo activate --latest)\n")
                f.write("dodo check-config-version\n")
                f.write("dodo check-version\n")
        if status == "off" and os.path.exists(script):
            os.unlink(script)



# -*- coding: utf-8 -*-

"""Utilities."""
from six.moves import input as raw_input
import os
import sys


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


_global_config = """
[DodoCommands]
projects_dir=~/projects
python_interpreter=python
diff_tool=diff
"""


def create_global_config():
    """Create config file and default_commands dir."""
    base_dir = os.path.expanduser('~/.dodo_commands')
    if not os.path.exists(base_dir):
        os.mkdir(base_dir)

    config_filename = os.path.join(base_dir, "config")
    if not os.path.exists(config_filename):
        with open(config_filename, 'w') as f:
            f.write(_global_config)

    default_commands_dir = os.path.join(base_dir, "default_commands")
    if not os.path.exists(default_commands_dir):
        os.mkdir(default_commands_dir)

    init_py = os.path.join(default_commands_dir, "__init__.py")
    if not os.path.exists(init_py):
        with open(init_py, 'w') as f:
            pass


def remove_trailing_dashes(args):
    """Removes first -- item from args."""
    return args[1:] if args[:1] == ['--'] else args


def bordered(text):
    lines = text.splitlines()
    width = max(len(s) for s in lines)
    res = ['┌' + '─' * width + '┐']
    for s in lines:
        res.append('│' + (s + ' ' * width)[:width] + '│')
    res.append('└' + '─' * width + '┘')
    return '\n'.join(res) + '\n'



class Decorator:  # noqa
    def add_arguments(self, decorated, parser):  # noqa
        parser.add_argument(
            '--use-debugger',
            action='store_true',
            default=False,
            help="Run the command through the debugger"
        )

    def handle(self, decorated, use_debugger, **kwargs):  # noqa
        decorated.opt_use_debugger = use_debugger

    def modify_args(self, decorated, args, cwd):  # noqa
        if not decorated.opt_use_debugger:
            return args, cwd

        new_args = [decorated.get_config('/BUILD/debugger')] + args
        return new_args, cwd



"""Print the command line to cd to a folder inside the project folder."""
from . import DodoCommand
import sys


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            'to',
            nargs='?',
            help=(
                'cd to /ROOT/<to>_dir. For example: dodo cd src ' +
                'cds to the value of /ROOT/src_dir.')
        )

    def handle_imp(self, to, **kwargs):
        path = self.get_config("/ROOT/%s_dir" % (to or 'project'))
        sys.stdout.write("cd %s\n" % path)



"""This command opens a bash shell in the docker container."""

from . import DodoCommand


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('command', nargs='?')

    def handle_imp(self, command, **kwargs):  # noqa
        self.runcmd(
            ["/bin/bash"] + (["-c", command] if command else []),
            cwd=self.get_config("/DOCKER/default_cwd", None))



"""Pull the latest version of the Dodo Commands system."""

from dodo_commands.system_commands import DodoCommand, CommandError
from dodo_commands.framework.config import ConfigIO
import glob
import os
import sys


class Command(DodoCommand):  # noqa
    safe = False

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument(
            'src_dir',
            help="The src directory for the bootstrapped project"
        )
        parser.add_argument(
            'project_defaults_dir',
            help="Location relative to src_dir where the default project config is stored",
        )
        parser.add_argument('--force', dest="use_force", action="store_true")
        parser.add_argument(
            '--git-url', dest='git_url',
            help="Clone this repository to the src_dir location",
        )
        parser.add_argument(
            '--depth',
            type=int,
            default=0,
            help="Depth for cloning repositories"
        )
        parser.add_argument(
            '--branch',
            help="Branch to checkout after cloning"
        )

    def _report(self, msg):
        sys.stderr.write(msg + "\n")

    def _copy_defaults(self, project_defaults_dir, use_force):
        res_dir = os.path.join(self.project_dir, "dodo_commands", "res")
        for filename in glob.glob(os.path.join(project_defaults_dir, "*")):
            dest_path = os.path.join(res_dir, os.path.basename(filename))
            if os.path.exists(dest_path):
                if self.opt_confirm:
                    print("Warning, destination path already exists: %s" % dest_path)
                elif use_force:
                    print("Overwriting existing path: %s" % dest_path)
                else:
                    raise CommandError(
                        "Destination path %s already exists. " % dest_path +
                        "Use the --confirm or --force flag to overwrite it."
                    )
            self.runcmd(["cp", "-rf", filename, dest_path])

    def _clone(self, src_dir, git_url, depth, branch):
        if os.path.exists(src_dir):
            print("Path already exists, not cloning into: " + src_dir)
        else:
            self.runcmd(
                [
                    "git",
                    "clone",
                    git_url,
                    os.path.basename(src_dir)
                ] + (["--depth", depth] if depth else []),
                cwd=os.path.dirname(src_dir)
            )
            if branch:
                self.runcmd(
                    [
                        "git",
                        "checkout",
                        branch
                    ],
                    cwd=src_dir
                )

        return True

    def _create_symlink(self, project_defaults_dir):
        target_dir = os.path.join(
            self.project_dir,
            "dodo_commands/default_project"
        )
        if os.path.exists(target_dir):
            self.runcmd(["rm", target_dir])
        self.runcmd(["ln", "-s", project_defaults_dir, target_dir])

    def _write_src_dir(self, src_dir):
        config = ConfigIO().load(load_layers=False)
        config['ROOT']['src_dir'] = src_dir
        ConfigIO().save(config)

    def handle_imp(
        self,
        src_dir,
        project_defaults_dir,
        use_force,
        git_url,
        depth,
        branch,
        **kwargs
    ):  # noqa
        self.project_dir = self.get_config('/ROOT/project_dir')
        clone_dir = (
            src_dir
            if os.path.isabs(src_dir) else
            os.path.join(self.project_dir, src_dir)
        )

        if git_url:
            self._clone(clone_dir, git_url, depth, branch)

        project_defaults_dir = os.path.join(clone_dir, project_defaults_dir)
        if not os.path.exists(project_defaults_dir):
            self._report(
                "Default project location %s not " +
                "found." % project_defaults_dir
            )
            return None

        self._create_symlink(project_defaults_dir)

        if git_url:
            self._copy_defaults(project_defaults_dir, use_force)

        self._write_src_dir(
            src_dir
            if os.path.isabs(src_dir) else
            os.path.join("${/ROOT/project_dir}", src_dir)
        )



# noqa
from dodo_commands.system_commands import DodoCommand, CommandError
import sys
import os


def _extra_dir():
    import dodo_commands
    return os.path.join(
        os.path.dirname(dodo_commands.__file__), "extra"
    )


def _packages_in_extra_dir():
    extra_dir = _extra_dir()
    packages = [
        x for x in os.listdir(extra_dir)
        if os.path.isdir(os.path.join(extra_dir, x))
        and not x.startswith('__')
    ]

    if len(packages) == 0:
        return ""

    if len(packages) == 1:
        msg = " The %s package is found automagically " % packages[0]
    else:
        packages[-1] = "and " + packages[-1]
        msg = " The %s packages are found automagically " % ", ".join(packages)

    return (
        msg + " in the dodo_commands.extra package" +
        ", e.g. the following works: dodo install-default-commands %s." % packages[0]
    )


class Command(DodoCommand):  # noqa
    help = (
        "Install default command directories supplied by the " +
        "paths argument: dodo install-default-commands " +
        "/path/to/my/commands. " + _packages_in_extra_dir()
    )
    safe = False

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument(
            "paths",
            nargs='*',
            help='Create symlinks to these command directories'
        )
        parser.add_argument(
            "--pip",
            nargs='*',
            help='Pip install the commands in these packages'
        )

    def _report_error(self, msg):
        sys.stderr.write(msg + os.linesep)

    def _install_commands(self, path):
        """Install the dir with the default commands."""
        dest_dir = os.path.join(
            os.path.expanduser("~/.dodo_commands/default_commands"),
            os.path.basename(path)
        )

        if not os.path.exists(path):
            alt_path = os.path.join(_extra_dir(), path)
            if os.path.exists(alt_path):
                path = alt_path
            else:
                self._report_error("Error: path not found: %s" % path)
                return False

        if os.path.exists(dest_dir):
            return False

        try:
            os.symlink(os.path.abspath(path), dest_dir)
        except:
            self._report_error("Error: could not create a symlink in %s." % dest_dir)
            return False

        return True

    def _install_package(self, package):
        pip = os.path.join(os.path.dirname(sys.executable), "pip")
        if pip.startswith('/usr/') and not os.path.exists(pip):
            alt_pip = pip.replace("/usr/", "/usr/local/")
            if os.path.exists(alt_pip):
                pip = alt_pip
            else:
                raise CommandError(
                    "Expected to find a pip executable at location %s or %s."
                    % (pip, alt_pip)
                )

        default_commands_dir = os.path.expanduser("~/.dodo_commands/default_commands")
        self.runcmd([
            pip, 'install', '--upgrade', '--target', default_commands_dir, package
        ])

    def handle_imp(self, paths, pip, **kwargs):  # noqa
        import dodo_commands
        dodo_commands_path = dodo_commands.__path__[0]
        if os.path.realpath(dodo_commands_path) != dodo_commands_path:
            raise CommandError("Please deactivate your dodo project first by running 'deactivate'.")

        if paths:
            for path in paths:
                self._install_commands(path)

        if pip:
            for package in pip:
                self._install_package(package)



"""Package default_commands."""
import fnmatch
import inspect
import os

from dodo_commands.framework import (BaseCommand, CommandPath)
from dodo_commands.framework.command_error import CommandError

from importlib import import_module
from plumbum import FG, ProcessExecutionError, local
from dodo_commands.framework.util import query_yes_no


class DodoCommand(BaseCommand):  # noqa
    safe = True
    _loaded_decorators = None

    def all_decorators(self):
        """Returns a mapping from decorator name to its directory."""
        command_path = CommandPath(self.config)
        command_path.extend_sys_path()
        result = {}
        for item in command_path.items:
            try:
                module_path = os.path.join(item.module_path, "decorators")
                module = import_module(module_path.replace("/", "."))
                for decorator in os.listdir(module.__path__[0]):
                    name, ext = os.path.splitext(decorator)
                    if ext == '.py' and name != '__init__':
                        result[name] = module_path
            except ImportError:
                continue
        return result

    @staticmethod
    def _load_decorator(name, directory):
        """Load and return decorator class in module with given name."""
        module = import_module(directory.replace("/", ".") + "." + name)
        return module.Decorator()

    def _command_name(self):
        script_filename = os.path.basename(inspect.getfile(self.__class__))
        return os.path.splitext(script_filename)[0]

    def _uses_decorator(self, decorator_name):
        patterns = (
            self.config.get('ROOT', {}).get('decorators', {}).get(decorator_name, [])
        )
        command_name = self._command_name()
        approved = [
            pattern for pattern in patterns if
            not pattern.startswith("!") and fnmatch.filter([command_name], pattern)
        ]
        rejected = [
            pattern for pattern in patterns if
            pattern.startswith("!") and fnmatch.filter([command_name], pattern[1:])
        ]
        return len(approved) and not len(rejected)

    def _get_decorators(self):
        if self._loaded_decorators is None:
            self._loaded_decorators = [
                self._load_decorator(name, directory)
                for name, directory in self.all_decorators().items()
                if self._uses_decorator(name)
            ]
        return self._loaded_decorators

    def add_arguments(self, parser):
        """Entry point for subclassed commands to add custom arguments."""
        parser.add_argument(
            '--confirm',
            action='store_true',
            help="Confirm each performed action before its execution"
        )

        parser.add_argument(
            '--echo',
            action='store_true',
            help="Print all commands instead of running them"
        )

        for decorator in self._get_decorators():
            decorator.add_arguments(self, parser)

        self.add_arguments_imp(parser)

    def add_arguments_imp(self, parser):  # noqa
        pass

    def handle(  # noqa
        self,
        confirm=False,
        echo=False,
        **kwargs
    ):
        self.opt_confirm = confirm
        self.opt_echo = echo

        if echo and not self.safe:
            raise CommandError(
                "The --echo option is not supported for unsafe commands.\n" +
                "Since this command is marked as unsafe, some operations will " +
                "be performed directly, instead of echoed to the console. " +
                "Use --confirm if you wish to execute with visual feedback. "
            )

        if confirm and not self.safe:
            if not query_yes_no(
                "Since this command is marked as unsafe, some operations will " +
                "be performed without asking for confirmation. Continue?",
                default="no"
            ):
                return

        for decorator in self._get_decorators():
            decorator.handle(self, **kwargs)
        self.handle_imp(**kwargs)

    def handle_imp(self, **kwargs):  # noqa
        raise CommandError("Not implemented")

    def runcmd(self, args, cwd=None):
        for decorator in self._get_decorators():
            args, cwd = decorator.modify_args(self, args, cwd)

        func = local[args[0]][args[1:]]
        if self.opt_echo:
            print(func)
            return False

        if self.opt_confirm:
            print("(%s) %s" % (cwd or local.cwd, func))
            if not query_yes_no("continue?"):
                return False
            print()

        if cwd and not os.path.exists(cwd):
            raise CommandError("Directory not found: %s" % cwd)

        with local.cwd(cwd or local.cwd):
            try:
                variable_map = self.get_config('/ENVIRONMENT/variable_map', {})
                with local.env(**variable_map):
                    func & FG
                return True
            except ProcessExecutionError:
                print("\nDodo Commands error while running this command:\n\n%s" % func)
                return False



from dodo_commands.framework.command_error import CommandError
import os
import re


class KeyNotFound(CommandError):
    """Raised if an xpath is not found in a configuration."""

    pass


class Key:
    """Access a nested value in a dict."""

    def __init__(self, config, xpath):  # noqa
        self.config = config
        # the list of sub-keys along the path to an item in self.config
        self.xpath = xpath

    def child(self, subkey):
        """Return key for child with subkey."""
        return Key(self.config, list(self.xpath) + [subkey])

    def _step_into(self, node, subkey):
        try:
            if isinstance(node, type(dict())):
                return node[subkey]
            if isinstance(node, type(list())):
                return node[int(subkey)]
        except:
            raise KeyNotFound("Cannot get value at %s\n" % self)

    def get(self):  # noqa
        node = self.config
        for subkey in self.xpath:
            node = self._step_into(node, subkey)
        return node

    def set(self, value):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        node[self.xpath[-1]] = value

    def remove(self):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        del node[self.xpath[-1]]

    def __repr__(self):  # noqa
        return str(self.xpath)


class ConfigExpander:
    """Expand environment variables and references in the config."""

    _regexp = r"\$\{(/[^\}]*)\}"

    @classmethod
    def _is_fully_expanded(cls, key, processed_xpaths):
        if key.xpath in processed_xpaths:
            return True

        value = key.get()
        if cls._is_leaf(value):
            return False

        return all([
            cls._is_fully_expanded(key.child(x), processed_xpaths)
            for x in cls._subkeys_of(value)
        ])

    @classmethod
    def _must_evaluate(cls, key):
        last_subkey = key.xpath[-1]
        return (
            isinstance(last_subkey, type(str())) and
            last_subkey.endswith("_EVAL")
        )

    @classmethod
    def _evaluate(cls, key):
        value = key.get()
        if cls._is_leaf(value):
            try:
                key.set(eval(value))
            except:
                raise CommandError(
                    "Cannot evaluate value '%s' for key %s\n" % (value, key)
                )
        else:
            for k in cls._subkeys_of(value):
                if cls._is_leaf(value[k]):
                    cls._evaluate(key.child(k))

    @classmethod
    def _is_leaf(cls, value):
        return not(
            isinstance(value, type(dict())) or
            isinstance(value, type(list()))
        )

    @classmethod
    def _subkeys_of(cls, value):
        if isinstance(value, type(dict())):
            return [k for k in value]
        if isinstance(value, type(list())):
            return range(len(value))
        return []

    @classmethod
    def _key_for_match(cls, matched_str, config):
        key = Key(config, [])
        for index in [ix for ix in matched_str.split("/") if ix]:
            if isinstance(key.get(), type(list())):
                subkey = int(index)
            else:
                subkey = index
            key = key.child(subkey)
        return key

    @classmethod
    def _expand(cls, value, config, processed_xpaths):
        matches = re.finditer(cls._regexp, value)
        for match in reversed([m for m in matches]):
            replacement = cls._key_for_match(match.group(1), config)
            if replacement.xpath not in processed_xpaths:
                return False, None
            value = (
                value[:match.start()] +
                str(replacement.get()) +
                value[match.end():]
            )
        return True, os.path.expandvars(value)

    @classmethod
    def _expand_value_for(cls, key, processed_xpaths):
        value = key.get()
        if not cls._is_leaf(value):
            return cls._is_fully_expanded(key, processed_xpaths), None
        if not isinstance(value, type(str())):
            return True, None
        return cls._expand(
            value, key.config, processed_xpaths
        )

    @classmethod
    def _for_each_xpath(cls, config, node, f, xpath=None):
        if xpath is None:
            xpath = []

        for subkey in cls._subkeys_of(node):
            value = node[subkey]
            f(config, list(xpath) + [subkey])
            cls._for_each_xpath(
                config, value, f, list(xpath) + [subkey]
            )

    @classmethod
    def _expanded_key(cls, key, processed_xpaths):
        """Return key with variable expansion on last subkey.

        Return None if not all variables could be expanded.
        """
        last_subkey = key.xpath[-1]
        if not isinstance(last_subkey, type(str())):
            return key

        is_expanded, expanded_last_subkey = cls._expand(
            last_subkey, key.config, processed_xpaths
        )
        if not is_expanded:
            return None
        return Key(
            key.config, key.xpath[:-1] + [expanded_last_subkey]
        )

    @classmethod
    def run(cls, config):  # noqa
        all_keys = []
        cls._for_each_xpath(
            config, config,
            lambda config, xpath: all_keys.append(Key(config, xpath))
        )

        processed_xpaths = []
        changed = True
        while all_keys and changed:
            changed = False
            for key in list(all_keys):
                expanded_key = cls._expanded_key(key, processed_xpaths)
                if expanded_key is None:
                    continue

                is_expanded, expanded_value = cls._expand_value_for(
                    key, processed_xpaths
                )

                if is_expanded:
                    if expanded_value is not None:
                        expanded_key.set(expanded_value)
                        if expanded_key.xpath != key.xpath:
                            key.remove()
                    if cls._must_evaluate(expanded_key):
                        cls._evaluate(expanded_key)
                    processed_xpaths.append(expanded_key.xpath)
                    all_keys.remove(key)
                    changed = True

        if all_keys:
            print("Warning: unexpanded keys %s" % all_keys)



# This file is based on source code from the Django Software Foundation,
# using the following license:

# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification,          # noqa
# are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.

#     3. Neither the name of Django nor the names of its contributors may be used           # noqa
#        to endorse or promote products derived from this software without
#        specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND           # noqa
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR           # noqa
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES            # noqa
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON            # noqa
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Root of the framework module."""

import argcomplete
from argparse import ArgumentParser
from collections import defaultdict
import ruamel.yaml
import os
import pkgutil
import subprocess
import sys
import traceback
from importlib import import_module

from dodo_commands.framework.config import (
    CommandPath, get_project_dir, load_dodo_config
)
from dodo_commands.framework.base import (
    BaseCommand, CommandError
)


def get_version():  # noqa
    return "0.9.1"


def find_commands(module_dir):
    """
    Return a list of all the command names.

    Given a path to a dodo_commands directory, returns a list of all the
    command names that are available.

    Returns an empty list if no commands are defined.
    """
    return [
        name for _, name, is_pkg in pkgutil.iter_modules([module_dir])
        if not is_pkg and not name.startswith('_')
    ]


def load_command_class(module_dir, name):
    """
    Return the Command class instance for module_dir and name.

    Given a command name and module directory, returns the Command
    class instance. All errors raised by the import process
    (ImportError, AttributeError) are allowed to propagate.
    """
    def install_packages(meta_data_filename):
        """Pip install packages found in meta_data_filename."""
        with open(meta_data_filename) as f:
            meta_data = ruamel.yaml.round_trip_load(f.read())
            print("\n--- Installing from %s ---" % meta_data_filename)
            pip = os.path.join(
                get_project_dir(),
                "dodo_commands/env/bin/pip"
            )
            subprocess.check_call(
                [pip, "install"] + meta_data['requirements'])
            print("--- Done ---\n\n")

    package_path = module_dir.replace("/", ".")
    import_path = '%s.%s' % (package_path, name)
    if module_dir in ("", None, "."):
        import_path = name

    try:
        module = import_module(import_path)
    except ImportError as e:
        try:
            base_path = import_module(package_path).__path__[0]
            meta_data_filename = os.path.join(base_path, name + ".meta")
            if os.path.exists(meta_data_filename):
                install_packages(meta_data_filename)
                module = import_module(import_path)
            else:
                print(traceback.print_exc(e))
                sys.exit(1)
        except ImportError as e:
            print(traceback.print_exc(e))
            sys.exit(1)

    return module.Command()


def get_commands():
    """
    Return a dictionary mapping command names to their Python module directory.

    The dictionary is in the format {command_name: module_name}. Key-value
    pairs from this dictionary can then be used in calls to
    load_command_class(module_name, command_name)
    """
    commands = {}
    command_path = CommandPath(load_dodo_config())
    command_path.extend_sys_path()
    for item in command_path.items:
        for command in find_commands(item.full_path):
            commands[command] = item.module_path
    return commands


def call_command(name, *args, **options):
    """
    Call the given command, with the given options and args/kwargs.

    This is the primary API you should use for calling specific commands.
    """
    # Load the command object.
    try:
        module_name = get_commands()[name]
    except KeyError:
        raise CommandError("Unknown command: %r" % name)

    if isinstance(module_name, BaseCommand):
        # If the command is already loaded, use it directly.
        command = module_name
    else:
        command = load_command_class(module_name, name)

    # Simulate argument parsing to get the option defaults.
    parser = command.create_parser('', name)
    # Use the `dest` option name from the parser option
    opt_mapping = {
        sorted(
            s_opt.option_strings
        )[0].lstrip('-').replace('-', '_'): s_opt.dest
        for s_opt in parser._actions if s_opt.option_strings
    }
    arg_options = {
        opt_mapping.get(key, key): value
        for key, value in options.items()
    }
    defaults = parser.parse_args(args=args)
    defaults = dict(defaults._get_kwargs(), **arg_options)
    # Move positional args out of options to mimic legacy optparse
    args = defaults.pop('args', ())

    return command.handle(*args, **defaults)


class ManagementUtility(object):
    """Internal helper class for executing commands."""

    def __init__(self, argv):  # noqa
        self.argv = argv
        self.prog_name = os.path.basename(self.argv[0])

    def main_help_text(self, commands_only=False, commands=None):
        """Return the script's main help text, as a string."""
        if commands is None:
            commands = get_commands()

        if commands_only:
            usage = sorted(commands.keys())
        else:
            usage = [
                "",
                "Version %s (%s --version)." % (
                    get_version(), self.prog_name
                ),
                "Type '%s help <subcommand>' for help on "
                "a specific subcommand." % self.prog_name,
                "Available subcommands (dodo help --commands):",
            ]
            commands_dict = defaultdict(lambda: [])
            for name, app in commands.items():
                app = app.rpartition('.')[-1]
                commands_dict[app].append(name)
            for app in sorted(commands_dict.keys()):
                usage.append("")
                for name in sorted(commands_dict[app]):
                    usage.append("    %s" % name)

        return '\n'.join(usage)

    def fetch_command(self, subcommand, commands=None):
        """
        Fetch the given subcommand.

        Tries to fetch the given subcommand, printing a message with the
        appropriate command called from the command line if it can't be found.
        """
        # Get commands outside of try block to prevent swallowing exceptions
        if commands is None:
            commands = get_commands()

        if subcommand not in commands:
            print("Unknown dodo command: %s" % subcommand)
            sys.exit(1)

        module_name = commands[subcommand]
        if isinstance(module_name, BaseCommand):
            # If the command is already loaded, use it directly.
            klass = module_name
        else:
            klass = load_command_class(module_name, subcommand)
        return klass

    def execute(self):
        """
        Execute subcommand.

        Given the command-line arguments, this figures out which subcommand is
        being run, creates a parser appropriate to that command, and runs it.
        """
        commands = get_commands()

        if "_ARGCOMPLETE" in os.environ:
            words = os.environ['COMP_LINE'].split()
            subcommand = words[1]

            if subcommand not in commands:
                parser = ArgumentParser()
                parser.add_argument(
                    'command',
                    choices=[x for x in commands.keys()]
                )
                argcomplete.autocomplete(parser)

            os.environ['COMP_LINE'] = ' '.join(words[:1] + words[2:])
            os.environ['COMP_POINT'] = str(
                int(os.environ['COMP_POINT']) - (len(subcommand) + 1)
            )
        else:
            try:
                subcommand = self.argv[1]
            except IndexError:
                subcommand = 'help'  # Display help if no arguments were given.

        args = self.argv[2:]

        if subcommand == '--version':
            sys.stdout.write(get_version() + '\n')
        elif subcommand == 'help':
            if '--commands' in args:
                sys.stdout.write(
                    self.main_help_text(
                        commands_only=True, commands=commands
                    ) + '\n'
                )
            elif len(args) < 1:
                sys.stdout.write(self.main_help_text() + '\n')
            else:
                self.fetch_command(
                    args[0], commands
                ).print_help(self.prog_name, args[0])
        else:
            self.fetch_command(subcommand, commands).run_from_argv(
                self.argv, subcommand
            )


def execute_from_command_line(argv):
    """A simple method that runs a ManagementUtility."""
    utility = ManagementUtility(argv)
    utility.execute()



"""Build a docker image."""

import os

from . import DodoCommand
from dodo_commands.framework import CommandError


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument(
            '--image',
            dest="docker_image",
            help=(
                "Identifies the docker image. "
                "You should supply a value foo:bar. "
                "A file Dockerfile.foo.bar will be used "
                "as the input docker file."
            )
        )

    def _copy_extra_dirs(self, local_dir, extra_dirs):
        for extra_dir_name, extra_dir_path in extra_dirs.items():
            local_path = os.path.join(local_dir, extra_dir_name)
            if os.path.exists(local_path):
                raise CommandError(
                    "Cannot copy to existing path: " + local_path
                )
            if not os.path.exists(extra_dir_path):
                raise CommandError(
                    "Cannot copy from non-existing path: " + extra_dir_path
                )
            self.runcmd(["cp", "-rf", extra_dir_path, local_path])

    def _remove_extra_dirs(self, local_dir, extra_dirs):
        for extra_dir_name, extra_dir_path in extra_dirs.items():
            local_path = os.path.join(local_dir, extra_dir_name)
            self.runcmd(["rm", "-rf", local_path])

    def handle_imp(self, docker_image, **kwargs):  # noqa
        res_dir = os.path.join(
            self.get_config("/ROOT/project_dir"), "dodo_commands", "res"
        )
        local_dir = self.get_config("/DOCKER/build_dir", res_dir)

        if not docker_image:
            docker_image = self.get_config("/DOCKER/image")

        extra_dirs = self.get_config("/DOCKER/extra_dirs", {})
        self._copy_extra_dirs(local_dir, extra_dirs)

        try:
            self.runcmd(
                [
                    "docker",
                    "build",
                    "-t",
                    docker_image,
                    "-f",
                    "Dockerfile.%s.%s" % tuple(docker_image.split(":", 1)),
                ] +
                [
                    ".",
                ],
                cwd=local_dir
            )
        finally:
            self._remove_extra_dirs(local_dir, extra_dirs)



"""Module for working with the dodo configurations."""

from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config_expander import ConfigExpander
from dodo_commands.framework.config_expander import Key, KeyNotFound  # noqa
import glob
import os
import sys
import ruamel.yaml


def get_project_dir():
    """Return the root dir of the current project."""
    return os.environ['DODO_COMMANDS_PROJECT_DIR']


def _default_config_base_dir(project_dir):
    return os.path.join(project_dir, "dodo_commands", "res")


class ConfigIO:
    """Read and write the dodo configuration."""

    def __init__(self, config_base_dir=None):
        """Arg config_base_dir is where config files are searched.

        Arg config_base_dir defaults to get_project_dir()/dodo_commands/res.
        """
        self.config_base_dir = (
            _default_config_base_dir(get_project_dir())
            if config_base_dir is None else
            config_base_dir
        )

    def _path_to(self, post_fix_paths):
        """Return path composed of config_base_dir and post_fix_paths list."""
        return os.path.join(self.config_base_dir, *post_fix_paths)

    def _layers(self, config):
        """Returns list of layer filenames"""
        result = []
        root = config.get('ROOT', {})
        layer_dir = root.get('layer_dir', '')
        for layer_filename in root.get('layers', []):
            pattern = os.path.expanduser(
                layer_filename
                if os.path.dirname(layer_filename) else
                self._path_to([layer_dir, layer_filename])
            )
            result.extend(x for x in glob.glob(pattern))
        return result

    def _add_layer_to_config(self, config, layer):
        for section in (layer or {}):
            if section not in config:
                config[section] = layer[section]
                continue
            for key, value in layer[section].items():
                has_key = key in config[section]
                if has_key and isinstance(value, type(dict())):
                    config[section][key].update(value)
                elif has_key and isinstance(value, type(list())):
                    config[section][key].extend(value)
                else:
                    config[section][key] = value

    def _load_layers(self, config):
        for layer_filename in self._layers(config):
            if os.path.exists(layer_filename):
                self._add_layer_to_config(
                    config, self.load(layer_filename, load_layers=False)
                )
            else:
                print("Warning: layer not found: %s" % layer_filename)

    def load(self, config_filename='config.yaml', load_layers=True):
        """Get configuration."""
        if not get_project_dir():
            return None

        full_config_filename = self._path_to([config_filename])
        if not os.path.exists(full_config_filename):
            return None

        with open(full_config_filename) as f:
            config = ruamel.yaml.round_trip_load(f.read())
        if load_layers:
            self._load_layers(config)

        return config

    def save(self, config, config_filename='config.yaml'):
        """Write config to config_filename."""
        with open(self._path_to([config_filename]), 'w') as f:
            return f.write(ruamel.yaml.round_trip_dump(config))


class ConfigLoader:
    """Load the project's dodo config and expand it."""
    def _add_to_config(self, config, section, key, value):
        if section in config:
            if key not in config[section]:
                config[section][key] = value

    def _system_commands_dir(self):
        """Return directory where system command scripts are stored"""
        import dodo_commands.system_commands
        return os.path.dirname(dodo_commands.system_commands.__file__)

    def _extend_config(self, config):
        """Add special values to the project's config"""
        project_dir = get_project_dir()
        if project_dir:
            self._add_to_config(
                config, "ROOT", "project_name", os.path.basename(project_dir))
            self._add_to_config(
                config, "ROOT", "project_dir", project_dir)
            self._add_to_config(
                config,
                "ROOT",
                "res_dir",
                _default_config_base_dir(project_dir)
            )

    def _extend_command_path(self, config):
        """Add the system commands to the command path"""
        self._add_to_config(config, "ROOT", "command_path", [])
        config['ROOT']['command_path'].append(
            [os.path.dirname(self._system_commands_dir()), "system_commands"]
        )

    def load(self, config_base_dir=None):
        config = ConfigIO(config_base_dir).load() or dict(ROOT={})
        self._extend_command_path(config)
        self._extend_config(config)
        ConfigExpander().run(config)
        return config


def load_dodo_config(config_base_dir=None):
    return ConfigLoader().load(config_base_dir)


class CommandPath:
    """Read search paths for command scripts from the configuration."""

    class Item:  # noqa
        sys_path = ""
        module_path = ""

        @property
        def full_path(self):  # noqa
            return os.path.join(self.sys_path, self.module_path)

    def __init__(self, config):
        """config_base_dir is the directory where config files are searched.

        The config_base_dir arg defaults to_
        _default_project_base_dir(project_dir).
        """
        exclude_patterns = self._exclude_patterns(config)
        excluded_command_dirs = [
            x.full_path for x in self._create_items(exclude_patterns)
        ]

        include_patterns = self._include_patterns(config)
        self.items = [
            x for x in self._create_items(include_patterns)
            if x.full_path not in excluded_command_dirs
        ]

    def _include_patterns(self, config):
        return config.get('ROOT', {}).get('command_path', [])

    def _exclude_patterns(self, config):
        return config.get('ROOT', {}).get('command_path_exclude', [])

    def _create_items(self, patterns):
        items = []
        for pattern in patterns:
            sys_path, module_path = self._split_pattern(pattern)
            for command_dir in self._glob_command_dirs(sys_path, module_path):
                items.append(self._create_item(sys_path, command_dir))
        return items

    def _create_item(self, sys_path, command_dir):
        item = CommandPath.Item()
        item.sys_path = sys_path
        item.module_path = os.path.relpath(command_dir, sys_path)
        return item

    def _glob_command_dirs(self, sys_path, module_path):
        return [
            x for x in glob.glob(os.path.join(sys_path, module_path))
            if os.path.isdir(x)
        ]

    def _split_pattern(self, pattern):
        try:
            sys_path, module_path = pattern
            sys_path = os.path.expanduser(sys_path)
        except:
            raise CommandError(
                "Patterns in command_path should be of " +
                "type [<sys-path>, <module-path>]. " +
                "Failing pattern: %s" % pattern
            )
        return sys_path, module_path

    def extend_sys_path(self):  # noqa
        for item in self.items:
            if item.sys_path not in sys.path:
                sys.path.append(item.sys_path)



"""Decorates command lines with docker arguments."""

from plumbum import local


class Decorator:  # noqa
    @classmethod
    def _get_docker_volume_list(cls, get_config, prefix=""):
        volume_list = get_config('DOCKER/volume_list', [])
        volume_map = get_config('/DOCKER/volume_map', {})
        return (
            [prefix + "%s:%s" % (x, x) for x in volume_list] +
            [prefix + "%s:%s" % key_val for key_val in volume_map.items()]
        )

    @classmethod
    def _get_docker_volumes_from_list(cls, get_config, prefix=""):
        volumes_from_list = get_config('DOCKER/volumes_from_list', [])
        return (
            [prefix + "%s" % x for x in volumes_from_list]
        )

    @classmethod
    def _get_docker_variable_list(cls, get_config, prefix=""):
        variable_list = get_config('/DOCKER/variable_list', [])
        variable_map = get_config('/ENVIRONMENT/variable_map', {})
        variable_map.update(get_config('/DOCKER/variable_map', []))

        return (
            [prefix + "%s" % x for x in variable_list] +
            [
                prefix + "%s=%s" % key_val
                for key_val in variable_map.items()
            ]
        )

    @classmethod
    def _get_linked_container_list(cls, get_config, prefix=""):
        return [prefix + "%s" % x for x in get_config('/DOCKER/link_list', [])]

    @classmethod
    def get_docker_args(cls, get_config, extra_options):
        """
        Get docker args.

        The docker args precede the command which is run
        inside the docker.
        """
        return (
            [
                'run',
                '--rm',
            ] +
            extra_options +
            cls._get_docker_variable_list(get_config, '--env=') +
            cls._get_docker_volume_list(get_config, '--volume=') +
            cls._get_docker_volumes_from_list(get_config, '--volumes-from=') +
            cls._get_linked_container_list(get_config, '--link=') +
            get_config('/DOCKER/extra_options', []) +
            [
                get_config('/DOCKER/image'),
            ]
        )

    def add_arguments(self, decorated, parser):  # noqa
        parser.add_argument(
            '--non-interactive',
            action='store_true',
            help="Run docker calls without -i and -t"
        )

    def handle(self, decorated, non_interactive, **kwargs):  # noqa
        decorated.opt_non_interactive = non_interactive

    def modify_args(self, decorated, args, cwd):  # noqa
        is_enabled = decorated.get_config('/DOCKER/enabled', False)
        if is_enabled == "False" or not is_enabled:
            return args, cwd

        extra_options = []
        if not decorated.opt_non_interactive:
            extra_options.extend(['-i', '-t'])
        if cwd:
            extra_options.extend(['-w', cwd])
        if hasattr(decorated, "docker_options"):
            extra_options.extend(decorated.docker_options)

        new_args = (
            ['docker'] +
            self.get_docker_args(decorated.get_config, extra_options) +
            args
        )
        return new_args, local.cwd



"""Compare configuration version to version in original project config file."""
from . import DodoCommand
from dodo_commands.framework import get_version
from dodo_commands.framework.util import bordered
import os
import sys
import ruamel.yaml


class Command(DodoCommand):  # noqa
    def _get_version(self, config_filename):
        with open(config_filename) as f:
            config = ruamel.yaml.round_trip_load(f.read())
        version = config.get("ROOT", {}).get("required_dodo_commands_version", "").split(".")
        return [x for x in version if x != ""]

    def handle_imp(self, **kwargs):  # noqa
        config_filename = os.path.join(
            self.get_config("/ROOT/project_dir", ""),
            "dodo_commands",
            "res",
            "config.yaml"
        )

        required_version = self._get_version(config_filename)
        if required_version:
            actual_version = get_version().split(".")
            if required_version > actual_version:
                sys.stdout.write(bordered(
                    'The dodo_commands package needs to be upgraded (%s < %s). Tip: use "dodo upgrade"'
                    % (
                        ".".join(actual_version),
                        ".".join(required_version),
                    ),
                ))
                sys.stdout.write('\n')



"""Compare configuration version to version in original project config file."""
from . import DodoCommand
from dodo_commands.framework.util import bordered
import os
import sys
import ruamel.yaml


class Command(DodoCommand):  # noqa
    def _get_version(self, config_filename):
        with open(config_filename) as f:
            config = ruamel.yaml.round_trip_load(f.read())
        version = config.get("ROOT", {}).get("version", "").split(".")
        return [x for x in version if x != ""]

    def handle_imp(self, **kwargs):  # noqa
        project_dir = self.get_config("/ROOT/project_dir", "")
        original_file = os.path.join(
            project_dir, "dodo_commands", "default_project", "config.yaml"
        )
        if not os.path.exists(original_file):
            return

        original_version = self._get_version(original_file)
        if not original_version:
            sys.stderr.write(
                "No version found in original file %s\n" % original_file
            )
            return

        copied_file = os.path.join(project_dir, "dodo_commands", "res", "config.yaml")
        copied_version = self._get_version(copied_file)
        if not copied_version:
            sys.stderr.write(
                "No version found in user managed file %s\n" % copied_file
            )
            return

        if copied_version[:2] < original_version[:2]:
            sys.stdout.write(bordered(
                'Configuration needs update (%s < %s). Tip: use "dodo diff ."\n'
                % (
                    ".".join(copied_version),
                    ".".join(original_version),
                )
            ))
            sys.stdout.write('\n')



from setuptools import setup


setup(name='dodo_commands',
      version='0.9.1',
      description='Project-aware development environments, inspired by django-manage',
      url='https://github.com/mnieber/dodo_commands',
      download_url='https://github.com/mnieber/dodo_commands/tarball/0.9.1',
      author='Maarten Nieber',
      author_email='hallomaarten@yahoo.com',
      license='MIT',
      packages=[
          'dodo_commands',
          'dodo_commands.framework',
          'dodo_commands.system_commands',
      ],
      package_data={
          'dodo_commands': [
              'extra/__init__.py',
              'extra/git_commands/*.py',
              'extra/git_commands/*.meta',
              'extra/standard_commands/*.py',
              'extra/standard_commands/*.meta',
              'extra/standard_commands/decorators/*.py',
              'extra/webdev_commands/*.py',
              'extra/webdev_commands/*.meta',
          ]
      },
      entry_points={
          'console_scripts': [
              'dodo=dodo_commands.dodo:main',
          ]
      },
      install_requires=[
          'argcomplete',
          'plumbum',
          'ruamel.yaml',
          'six',
      ],
      zip_safe=False)



"""Show diff for a file in the Dodo Commands system directory."""

from . import DodoCommand
from six.moves import configparser
import os


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument(
            'file',
            help='Show diff for this file'
        )
        parser.add_argument(
            '--project-name',
            help='Compare to files from an alternative project'
        )

    def handle_imp(self, file, project_name, **kwargs):  # noqa
        project_dir = self.get_config("/ROOT/project_dir", "")

        config = configparser.ConfigParser()
        config.read(os.path.expanduser("~/.dodo_commands/config"))

        diff_tool = config.get("DodoCommands", "diff_tool")
        res_dir = os.path.join(project_dir, "dodo_commands", "res")

        if project_name:
            ref_project_dir = os.path.abspath(
                os.path.join(project_dir, "..", project_name)
            )
            original_file = os.path.join(
                ref_project_dir, "dodo_commands", "res", file
            )
        else:
            original_file = os.path.realpath(
                os.path.join(
                    project_dir, "dodo_commands", "default_project", file
                )
            )
        copied_file = os.path.join(res_dir, file)

        self.runcmd(
            [diff_tool, original_file, copied_file],
            cwd=res_dir
        )



"""Finds a directory or file inside the current project."""
from . import DodoCommand
from dodo_commands.framework.config import CommandPath
import os
import sys

script_src = """# noqa
from dodo_commands.system_commands import DodoCommand


class Command(DodoCommand):  # noqa
    help = ""

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument('foo')
        parser.add_argument(
            '--bar',
            required=True,
            help=''
        )

    def handle_imp(self, foo, bar, **kwargs):  # noqa
        self.runcmd(["echo"], cwd=".")
"""

class Command(DodoCommand):  # noqa
    help = "Creates a new Dodo command."
    safe = False

    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        parser.add_argument('name')
        parser.add_argument(
            '--next-to',
            required=True,
            help='Create the new command at the location of this command'
        )

    def handle_imp(self, name, next_to, **kwargs):  # noqa
        dest_path = None
        command_path = CommandPath(self.config)
        for item in command_path.items:
            script_path = os.path.join(
                item.full_path, next_to + ".py"
            )
            if os.path.exists(script_path):
                dest_path = os.path.join(
                    item.full_path, name + ".py"
                )

        if not dest_path:
            sys.stderr.write("Script not found: %s\n" % next_to)
            return

        with open(dest_path, "w") as f:
            f.write(script_src)

        print(dest_path)
